/* exp - expand environment variables and other stuff */

/*
input           output
-----           ------
%%              a single percent sign (%)
%n%             command line argument n
%include:f%     contents of file f
%date:fmt%      date using format fmt -- see datetimeformats for details
%dt%            date and time in standard format
%date%          date in standard format
%time%          time in standard format
%pwd%           current working directory
%!cmd%          output of command cmd executed using bash or system()
%VAR%           environment variable VAR if it exists or the empty string
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <time.h>
#include <unistd.h>

#define DELIMITER '%'
#define DATE_DELIMITER '$'
#define MAX_VAR_LEN 1024

void processvar(char *variable, int argc, char **argv);
int isnumeric(char *string);
void insert(FILE *fin);
void printtime(char *fmt);
void execute(char *command, int argc, char **argv); /* execute-(win|unix).c */

int main(int argc, char **argv)
{
   int c;
   char lastc;
   char variable[MAX_VAR_LEN + 1];
   int  varindex=0;
   int invar = 0;
   int eating = 0;

   while ((c = getchar()) != EOF)
   {
      if (c == DELIMITER)
      {
         if (invar && (lastc != '\\'))
         {
            invar = 0;
            variable[varindex] = '\0';
            if (!eating)
               processvar(variable, argc, argv);
            else
               eating = 0;
            varindex = 0;
         }
         else
         {
            invar = 1;
            lastc = '\0';
         }
      }
      else
      {
         if (invar)
         {
            if ((c == '\\') && (lastc == '\\'))
               continue;
            if (!eating)
            {
               if (varindex >= MAX_VAR_LEN)
                  eating=1, fprintf(stderr, "%s: Variables connot be longer "
                                    "than %d characters\n", argv[0],
                                    MAX_VAR_LEN);
               else
                  variable[varindex++] = c;
            }
            lastc = c;
         }
         else
            putchar(c);
      }
   }
   return 0;
}

void processvar(char *variable, int argc, char **argv)
{
   char *argument;
   char *string;
   int argnumber;
   FILE *fin;

   if (*variable == '\0')
      putchar(DELIMITER);
   else if (isnumeric(variable) && ((argnumber=atoi(variable)) < argc) &&
           (argnumber >= 0))
      printf("%s", argv[argnumber]);
   else if (strstr(variable,"include:")==variable)
   {
      argument = &variable[8];
      if (*argument == '\0')
         fprintf(stderr, "%s: Missing Filename in %cinclude:%c directive.\n",
                 argv[0], DELIMITER, DELIMITER);
      else if ((fin = fopen(argument,"r")) == NULL)
         fprintf(stderr, "%s: Error opening file '%s'.\n", argv[0], argument);
      else
      {
         insert(fin);
         if (fclose(fin) == EOF)
            fprintf(stderr, "%s: Error closing file '%s'.\n",
                    argv[0], argument);
      }
   }
   else if (strstr(variable, "date:")==variable)
   {
      argument = &variable[5];
      if (*argument == '\0')
         fprintf(stderr, "%s: Missing argument in %cdate:%c directive.\n",
                 argv[0], DELIMITER, DELIMITER);
      else
         printtime(argument);
   }
   else if (strstr(variable, "dt")==variable)
      printtime("$m/$d/$Y $h:$0n:$0s $AM");
   else if (strstr(variable, "date")==variable)
      printtime("$m/$d/$Y");
   else if (strstr(variable, "time")==variable)
      printtime("$h:$0n:$0s $AM");
   else if (strstr(variable, "pwd")==variable)
   {
      char *wd;
      wd = getcwd(NULL, 0);
      printf("%s", wd);
      free(wd);
   }
   else if (strstr(variable, "!")==variable)
   {
      argument = &variable[1];
      if (*argument == '\0')
         fprintf(stderr, "%s: Missing argument in %c!%c directive.\n",
                 argv[0], DELIMITER, DELIMITER);
      else
         execute(argument, argc, argv);
   }
   else
      printf("%s", ((string = getenv(variable)) == NULL) ? "" : string);
}

int isnumeric(char *string)
{
   while (*string != '\0')
      if (!isdigit(*string++))
         return 0;
   return 1;
}

void insert(FILE *fin)
{
   int c;
   while ((c=getc(fin)) != EOF)
      putchar(c);
   return;
}

void printtime(char *fmt)
{
   int mod;
   static char *month[] = {"January","February","March","April","May","June",
                           "July","August","September","October","November",
                           "December"};
   static char *amonth[] = {"Jan","Feb","Mar","Apr","May","Jun","Jul","Aug",
                            "Sep","Oct","Nov","Dec"};
   static char *weekday[] = {"Sunday","Monday","Tuesday","Wednesday",
                             "Thursday","Friday","Saturday"};
   static char *weekday2[] = {"Su","M","Tu","W","Th","F","Sa"};
   static char *weekday4[] = {"Sun","Mon","Tues","Wed","Thurs","Fri","Sat"};
   struct tm ctime;
   time_t temp;

   temp = time(NULL);
   ctime = *localtime(&temp);
   while (*fmt != '\0')
   {
      if (*fmt == DATE_DELIMITER)
      {
         if (isdigit(*(++fmt)))
            mod = (*fmt++) - '0';
         else
            mod = -1;
         switch (*fmt)
         {
         case DATE_DELIMITER:
            putchar(DATE_DELIMITER);
            break;
         case 'y':
            printf("%02d", ctime.tm_year % 100);
            break;
         case 'Y':
            printf("%d", ctime.tm_year + 1900);
            break;
         case 'm':
            switch(mod)
            {
            case -1:
               printf("%d", ctime.tm_mon + 1);
               break;
            case 0:
               printf("%02d", ctime.tm_mon + 1);
               break;
            case 1:
               printf("%s", amonth[ctime.tm_mon]);
               break;
            case 2:
            default:
               printf("%s", month[ctime.tm_mon]);
               break;
            }
            break;
         case 'd':
            switch(mod)
            {
            case -1:
               printf("%d", ctime.tm_mday);
               break;
            case 0:
            default:
               printf("%02d", ctime.tm_mday);
               break;
            }
            break;
         case 'D':
            switch(mod)
            {
            case -1:
               printf("%d", ctime.tm_yday + 1);
               break;
            case 0:
            default:
               printf("%03d", ctime.tm_yday + 1);
               break;
            }
            break;
         case 'w':
            switch(mod)
            {
            case -1:
               printf("%d", ctime.tm_wday + 1);
               break;
            case 0:
               printf("%02d", ctime.tm_wday + 1);
               break;
            case 1:
               printf("%c", weekday[ctime.tm_wday][0]);
               break;
            case 2:
               printf("%s", weekday2[ctime.tm_wday]);
               break;
            case 3:
               printf("%.3s", weekday[ctime.tm_wday]);
               break;
            case 4:
               printf("%s", weekday4[ctime.tm_wday]);
               break;
            case 5:
            default:
               printf("%s", weekday[ctime.tm_wday]);
               break;
            }
            break;
         case 'W':
            switch(mod)
            {
            case -1:
               printf("%d", (ctime.tm_yday / 7) + 1);
               break;
            case 0:
            default:
               printf("%02d", (ctime.tm_yday / 7) + 1);
               break;
            }
            break;
         case 'h':
            switch(mod)
            {
            case -1:
               printf("%d", (ctime.tm_hour % 12) ? ctime.tm_hour % 12 : 12);
               break;
            case 0:
            default:
               printf("%02d", (ctime.tm_hour % 12) ? ctime.tm_hour % 12 : 12);
               break;
            }
            break;
         case 'H':
            switch(mod)
            {
            case -1:
               printf("%d", ctime.tm_hour);
               break;
            case 0:
            default:
               printf("%02d", ctime.tm_hour);
               break;
            }
            break;
         case 'n':
            switch(mod)
            {
            case -1:
               printf("%d", ctime.tm_min);
               break;
            case 0:
            default:
               printf("%02d", ctime.tm_min);
               break;
            }
            break;
         case 's':
            switch(mod)
            {
            case -1:
               printf("%d", ctime.tm_sec);
               break;
            case 0:
            default:
               printf("%02d", ctime.tm_sec);
               break;
            }
            break;
         case 'a':
            if (*(fmt+1) == 'm')
               fmt++, printf(ctime.tm_hour<11 ? "am" : "pm");
            else
               putchar(ctime.tm_hour<11 ? 'a' : 'p');
            break;
         case 'A':
            if (*(fmt+1) == 'M')
               fmt++, printf(ctime.tm_hour<11 ? "AM" : "PM");
            else
               putchar(ctime.tm_hour<11 ? 'A' : 'P');
            break;
         case 't':
            {
               struct tm utc;
               int timezone;
               utc = *gmtime(&temp);
               timezone = mktime(&utc) - mktime(&ctime);
            printf("%+05d", (timezone / 3600) * -100 - (timezone % 60));
                           /* the value is backwards; use - instead of + */
            }
            break;
         case 'T':
         {
            char timezonestr[51];
            strftime(timezonestr, 50, "%Z", &ctime);
            printf("%s", timezonestr);
            break;
         }
         default:
            putchar(*fmt);
            break;
         }
      }
      else
         putchar(*fmt);
      fmt++;
   }
}
